from typing import Sequence, Tuple
import click


@click.command('compile', help='Compile Configuration file')
@click.option('-c', '--configuration_file', required=True, help="Input configuration file.")
@click.option('-o', '--output_file', required=True, help="Output single configuration file.")
@click.option('--check/--nocheck', default=True, type=bool, help="If flag is TRUE compile procedure will stop if placeholders found.")
@click.option('--option', 'options', multiple=True, nargs=2, help='Bind dataset name to underfolder path')
def compile(configuration_file: str, output_file: str, check: bool, options: Sequence[Tuple[str, str]]):

    from choixe.configurations import XConfig
    import rich

    try:
        cfg = XConfig(filename=configuration_file)
    except Exception as e:
        rich.print(f"[red]Invalid configuration file: {e}[/red]")
        import sys
        sys.exit(1)

    placeholders = cfg.available_placeholders()
    for key, value in options:
        if key in placeholders:
            placeholder = placeholders[key]
            cfg.deep_set(key, placeholder.cast(value))

    if check:
        cfg.check_available_placeholders(close_app=True)

    cfg.save_to(output_file)



from choixe.importers import Importer


class TestImporters:

    def test_options(self):

        v = '@import(alpha.yml, beta.yml, delta.yml, default=gamma.yml)'

        importer = Importer.from_string(v)
        assert importer is not None

        print(importer.path)
        print(importer.options)
        print(importer.default_value)



from __future__ import annotations
from abc import ABCMeta

import typing
from schema import Schema


class MetaSpook(ABCMeta):
    # 😱😱😱😱😱😱😱😱😱😱😱😱😱😱
    SPOOKS_MAP = {}

    def __init__(self, name, bases, dict) -> None:
        MetaSpook.register_spook(self)

    @classmethod
    def register_spook(cls, x: typing.Type[Spook]):
        """ Register a Spook in the factory map

        :param x: target class to register. It must be a Spook!
        :type x: Type[Spook]
        """
        cls.SPOOKS_MAP[x.spook_name()] = x


class Spook(metaclass=MetaSpook):
    # 👻👻👻👻👻👻👻👻👻👻👻👻👻👻

    TYPE_FIELD = '__spook__'
    ARGS_FIELD = 'args'

    @classmethod
    def spook_name(cls) -> str:
        return cls.__name__

    @classmethod
    def full_spook_schema(cls) -> Schema:
        return Schema({
            Spook.TYPE_FIELD: cls.spook_name(),
            Spook.ARGS_FIELD: cls.spook_schema()
        })

    @classmethod
    def spook_schema(cls) -> typing.Union[None, dict]:
        return None

    @classmethod
    def from_dict(cls, d: dict):
        return cls(**d)

    def to_dict(self) -> dict:
        return self.__dict__

    @classmethod
    def _validate_schema(cls, d: dict):
        if cls.spook_schema() is not None:
            cls.full_spook_schema().validate(d)

    def serialize(self, validate: bool = True) -> dict:
        out = {
            Spook.TYPE_FIELD: self.spook_name(),
            Spook.ARGS_FIELD: self.to_dict()
        }
        if validate:
            self._validate_schema(out)
        return out

    @classmethod
    def hydrate(cls, d: dict, validate: bool = True) -> any:
        """ Hydrate dict in order to re-create the correspoding Spook

        :param d: serialized dict
        :type d: dict
        :param validate: TRUE to validate dict schema before hydratation, defaults to False
        :type validate: bool, optional
        :return: hydrated object
        :rtype: any
        """
        if validate:
            cls._validate_schema(d)
        return cls.from_dict(d[Spook.ARGS_FIELD])

    @classmethod
    def create(cls, d: dict, validate: bool = True) -> any:
        """ Creates an object starting from its serialized dict.
        It must be a Spook serialized class containing all deserialization attributes

        :param d: serialized dict
        :type d: dict
        :param validate: TRUE to validate dict schema before hydratation, defaults to False
        :type validate: bool, optional
        :raises RuntimeError: when corresponding class is not registered in the factory map
        :return: hydrated object
        :rtype: any
        """
        if d[Spook.TYPE_FIELD] in cls.SPOOKS_MAP:
            bt = cls.SPOOKS_MAP[d[Spook.TYPE_FIELD]]
            return bt.hydrate(d, validate=validate)
        raise RuntimeError(f'Non serializable data: {d}')



import inquirer
from choixe.placeholders import PlaceholderType
from choixe.configurations import XConfig
from choixe.inquirer import XInquirer


class TestInquirer(object):

    def test_xinquirer(self):

        assert XInquirer.safe_cast_placeholder(10, PlaceholderType.INT)
        assert XInquirer.safe_cast_placeholder(10, PlaceholderType.FLOAT)
        assert XInquirer.safe_cast_placeholder(10, PlaceholderType.STR)
        assert XInquirer.safe_cast_placeholder('10', PlaceholderType.INT)
        assert not XInquirer.safe_cast_placeholder('a string', PlaceholderType.INT)
        assert not XInquirer.safe_cast_placeholder('another string', PlaceholderType.FLOAT)
        assert XInquirer.safe_cast_placeholder(True, PlaceholderType.BOOL)
        assert not XInquirer.safe_cast_placeholder(0, PlaceholderType.BOOL)
        assert XInquirer.safe_cast_placeholder('hello', PlaceholderType.PATH)
        assert XInquirer.safe_cast_placeholder('what time is it?', PlaceholderType.DATE)

        d = {
            'a_int': '@int(a_int)',
            'another_int': '@int(a_int)',
            'a_float': '@float(a_float)',
            'a_bool': '@bool(a_bool)',
            'a_str1': '@str(a_str1)',
            'a_str2': '@str(a_str2,hello,hi)',
            'a_str3': '@str(a_str3,one,two,three,default=one)'
        }
        xconfig = XConfig.from_dict(d)
        placeholders = list(xconfig.available_placeholders().values())
        for ph in placeholders:
            q = XInquirer.placeholder_to_question(ph)
            if len(ph.options) > 0:
                assert isinstance(q, inquirer.List)
                assert len(q.choices) == len(ph.options)
            elif ph.type == PlaceholderType.BOOL:
                assert isinstance(q, inquirer.List)
                assert len(q.choices) == 2
            else:
                assert isinstance(q, inquirer.Text)
            assert q.name == ph.name
            assert q.default == ph.default_value

        unique_phs = XInquirer.unique_placeholders_with_order(xconfig.available_placeholders())
        assert len(unique_phs) == 6
        assert unique_phs[0].name == 'a_int'
        assert unique_phs[1].name == 'a_float'
        assert unique_phs[2].name == 'a_bool'
        assert unique_phs[3].name == 'a_str1'
        assert unique_phs[4].name == 'a_str2'
        assert unique_phs[5].name == 'a_str3'

    def test_prompt(self, monkeypatch):

        d = {
            'a_int': '@int(a_int)',
            'another_int': '@int(a_int)',
            'a_float': '@float(a_float)',
            'a_bool': '@bool(a_bool)',
            'a_str1': '@str(a_str1)',
            'a_str2': '@str(a_str2,hello,hi)',
            'a_str3': '@str(a_str3,one,two,three,default=one)'
        }

        fills = {
            'a_int': 2,
            'another_int': 666,
            'a_float': 2.22,
            'a_bool': True,
            'a_str1': 'hello',
            'a_str2': 'hi',
            'a_str3': 'three'
        }

        xconfig = XConfig.from_dict(d)

        import sys
        from io import StringIO

        old_prompt = XInquirer._system_prompt

        def custom_prompt(questions):
            return fills

        monkeypatch.setattr(XInquirer, '_system_prompt', custom_prompt, raising=True)
        new_xconfig = XInquirer.prompt(xconfig, close_app=False)

        assert len(new_xconfig.available_placeholders()) == 0



from click.testing import CliRunner
from choixe.cli.check import check
from pathlib import Path


def test_check(sample_configurations_data):
    runner = CliRunner()

    for cfg_data in sample_configurations_data:
        filename = cfg_data['filename']
        has_placeholders = cfg_data['has_placeholders']

        result = runner.invoke(check, ['-c', filename])
        assert result.exit_code == (1 if has_placeholders else 0)

        result = runner.invoke(check, ['-c', filename, '--noclose'])
        assert result.exit_code == 0


def test_check_wrong(sample_configurations_data):
    runner = CliRunner()

    for cfg_data in sample_configurations_data:
        filename = Path(str(cfg_data['filename']) + "@IMPOSSIBLE_ST3ING!")
        has_placeholders = cfg_data['has_placeholders']
        print(has_placeholders)

        result = runner.invoke(check, ['-c', filename])
        assert result.exit_code == 1




import re
from abc import ABC


class Directive(ABC):

    @classmethod
    def generate_directive_string(cls, label: str, args: list = None) -> str:
        raise NotImplementedError

    @classmethod
    def directive_pattern(cls) -> str:
        raise NotImplementedError()

    @classmethod
    def is_directive(cls, value: str) -> bool:
        matches = re.match(cls.directive_pattern(), str(value).strip())
        return matches is not None

    @classmethod
    def tokenize(cls, value: str) -> dict:
        raise NotImplementedError()

    @property
    def valid(self):
        return self._valid

    @property
    def label(self):
        return self._label

    @property
    def args(self):
        return self._args

    @property
    def kwargs(self):
        return self._kwargs

    @property
    def default_value(self):
        return self._default_value

    def __init__(self, value: str):
        if self.is_directive(value):
            self._valid = True
            self._tokens = self.tokenize(value)
            self._label = self._tokens['label'].lower()
            self._args = self._tokens['args']
            self._kwargs = self._tokens['kwargs']
            self._default_value = self._tokens['default_value']
        else:
            self._valid = False
            self._tokens = {}
            self._label = ''
            self._args = []
            self._kwargs = {}
            self._default_value = None


class DirectiveAT(Directive):
    DEFAULT_KEY = 'default'

    @classmethod
    def generate_directive_string(cls, label: str, args: list = None) -> str:
        if args is None:
            args = []
        return f'@{label}(' + ','.join(map(str, args)) + ')'

    @classmethod
    def directive_pattern(cls) -> str:
        return '[@].+[(](.+?)|[)]$'

    @classmethod
    def tokenize(cls, value: str) -> dict:
        value = value.strip()
        value = value.replace(' ', '')
        for ch in ['@', '(', ')']:
            value = value.replace(ch, ' ')
        values = value.split(' ')
        values = [x for x in values if len(x) > 0]
        assert len(values) >= 1

        label = values[0]
        args = values[1].split(',') if len(values) > 1 else []

        # default_token = f'{cls.DEFAULT_KEY}='

        kwargs = {}
        new_args = []
        for a in args:
            if '=' in a:
                key, value = [x.strip() for x in a.split('=')]
                kwargs[key] = value
            else:
                new_args.append(a)

        # # set default value if any
        default_value = None
        if cls.DEFAULT_KEY in kwargs:
            default_value = kwargs[cls.DEFAULT_KEY]

        return {
            'label': label,
            'args': new_args,
            'kwargs': kwargs,
            'default_value': default_value
        }


class DirectiveFactory(object):

    AVAILABLE_DIRECTIVES = [
        DirectiveAT
    ]

    @classmethod
    def build_directive_from_string(cls, value: str) -> Directive:
        for dtype in cls.AVAILABLE_DIRECTIVES:
            directive = dtype(value)
            if directive.valid:
                return directive
        return None


class DirectiveConsumer(object):

    def __init__(self, directive: Directive) -> None:
        self._directive = directive

    @property
    def directive(self):
        return self._directive

    @property
    def directive_class(self):
        return self._directive.__class__



import os
from choixe.directives import DirectiveAT
from choixe.placeholders import Placeholder
from typing import Sequence, Union
import copy
from box.box_list import BoxList
import numpy as np
from box.box import Box
from deepdiff import DeepDiff
from schema import Schema, Or, Regex, SchemaMissingKeyError, Use

import pydash
from pathlib import Path
import pytest
from choixe.configurations import XConfig


@pytest.fixture(scope="function")
def generic_temp_folder(tmpdir_factory):
    fn = tmpdir_factory.mktemp("_choixe_temp_folder")
    return fn


class PlaceholderGenerator:
    def __init__(self, base_name: str = 'v_') -> None:
        self.counter = 0
        self.base_name = base_name

    def create_placeholder(self, collection: Sequence, tp: str = None, others: Sequence = None):
        args = [f'{self.base_name}{self.counter}']
        if others is not None:
            args.extend(others)
        if tp is not None:
            if len(tp) > 0:
                p = DirectiveAT.generate_directive_string(tp, args)
                collection.append(p)
                return p

        # Default type
        p = DirectiveAT.generate_directive_string('str', args)
        collection.append(p)
        return p


def simple_data():
    sample_dict = {
        'one': 1,
        'two': np.random.randint(-50, 50, (3, 3)).tolist(),
        'three': {
            'a': 2,
            'c': True
        }
    }
    schema = Schema(
        {
            'one': Or(int, str),
            'two': list,
            'three': {
                'a': int,
                'c': bool
            }
        }
    )

    to_be_raplaced_keys = sorted([
        'three'
    ], reverse=True)

    return {
        'data': sample_dict,
        'schema': schema,
        'placeholders': [],
        'environment_variables': [],
        'to_be_replaced': to_be_raplaced_keys
    }


def complex_data():
    np.random.seed(666)

    placeholders = []
    env_variables = []
    pl_generator = PlaceholderGenerator()

    sample_dict = {
        'one': pl_generator.create_placeholder(placeholders, 'int', [6, 7, 8, 'default=10']),
        'two': np.random.randint(-50, 50, (3, 3)).tolist(),
        'three': {
            '3.1': 'TrueValue',
            '2.1': [False, False],
            'name': {
                'name_1': pl_generator.create_placeholder(placeholders, 'float'),
                'name_2': 2,
                'name_3': {
                    'this_is_a_list': [3.3, 3.3],
                    'this_is_A_dict': {
                        'a': {
                            'f': True,
                            's': False,
                            'numpy_array': np.array([1., 2, 3]),
                            'numpy_data': np.array([1., 2, 3])[0],
                            'tuple': ('a', 'b', '2'),
                            'boxlist': BoxList([1, 2, 3, 4]),
                            'boxdict': Box({'a': 2.2})
                        }
                    }
                }
            }
        },
        'first': {
            'f1': pl_generator.create_placeholder(placeholders, 'str', ['alpha', 'beta', 'gamma']),  # placeholders[2],
            'f2': 2.22,
            'f3': [3, 3, 3, 3, 3, 3],
            'external': {
                'ext': np.random.uniform(-2, 2, (2, 3)).tolist(),
                'ext_name': [pl_generator.create_placeholder(placeholders, 'bool')],
            },
            'ops': {
                'o1': pl_generator.create_placeholder(placeholders),  # placeholders[4],
                'o2': pl_generator.create_placeholder(placeholders, 'path'),  # placeholders[5],
                'o3': pl_generator.create_placeholder(placeholders, 'date'),  # placeholders[6],
                'o4': pl_generator.create_placeholder(placeholders, 'str', ['A', 'B', 'C', 'default=C']),  # placeholders[7],
                'o5': pl_generator.create_placeholder(placeholders),  # placeholders[8],
                'o6': pl_generator.create_placeholder(placeholders),  # placeholders[9]
            },
            'env': [
                pl_generator.create_placeholder(placeholders, 'env'),  # env_variables[0],
                pl_generator.create_placeholder(placeholders, 'env')
            ],
            'key': {
                'with': 120,
                'dots': pl_generator.create_placeholder(placeholders, 'object'),
            },
            'key.with.dots': pl_generator.create_placeholder(placeholders),  # placeholders[10],
            'key.with...many.....dots': pl_generator.create_placeholder(placeholders),
            'nested.key': {
                'with.dots': pl_generator.create_placeholder(placeholders)
            },
            'placeholder_object': pl_generator.create_placeholder(placeholders),
            'placeholder_cfg': pl_generator.create_placeholder(placeholders, 'cfg'),  # placeholders[15],
            'placeholder_cfg_root': pl_generator.create_placeholder(placeholders, 'cfg_root'),  # ,
        }
    }

    schema = Schema(
        {
            'one': Or(int, str),
            'two': list,
            'three': {
                '3.1': str,
                '2.1': [bool],
                'name': dict,
            },
            'first': {
                Regex(''): Or(str, int, list, float, dict)
            }
        }
    )

    to_be_raplaced_keys = sorted([
        'three.name',
        'three.name.name_3',
        'three.name.name_3.this_is_A_dict',
        'first',
        'first.external'
    ], reverse=True)

    return {
        'data': XConfig.decode(sample_dict),
        'schema': schema,
        'placeholders': placeholders,
        'environment_variables': env_variables,
        'to_be_replaced': to_be_raplaced_keys
    }


def data_to_test():
    return [
        simple_data(),
        complex_data()
    ]


def store_cfg(filename: Union[str, Path], d: dict):

    filename = str(filename)
    data = Box(d)
    if 'yml' in filename or 'yaml' in filename:
        Box(XConfig.decode(data.to_dict())).to_yaml(filename)
    if 'json' in filename:
        Box(XConfig.decode(data.to_dict())).to_json(filename)
    if 'toml' in filename:
        Box(XConfig.decode(data.to_dict())).to_toml(filename)


TESTABLE_ESTENSIONS = ['yaml', 'json', 'toml']


class TestXConfig(object):

    @pytest.mark.parametrize("cfg_extension", TESTABLE_ESTENSIONS)
    @pytest.mark.parametrize("data", data_to_test())
    def test_creation(self, generic_temp_folder, data, cfg_extension):  # TODO: toml 0.10.2 needed! toml has a bug otherwise

        generic_temp_folder = Path(generic_temp_folder)

        sample_dict = data['data']
        to_be_raplaced_keys = data['to_be_replaced']  # , _, _, to_be_raplaced_keys = data
        volatile_dict = copy.deepcopy(sample_dict)

        subtitutions_values = {}
        for k in to_be_raplaced_keys:
            random_name = k + f".{cfg_extension}"
            random_name = DirectiveAT.generate_directive_string('import', [random_name])
            subtitutions_values[random_name] = pydash.get(volatile_dict, k)

            pydash.set_(volatile_dict, k, random_name)

        output_cfg_filename = generic_temp_folder / f'out_config.{cfg_extension}'
        output_cfg_filename2 = generic_temp_folder / f'out_config2.{cfg_extension}'

        subtitutions_values[str(output_cfg_filename)] = volatile_dict

        saved_cfgs = []
        for directive_value, d in subtitutions_values.items():
            directive = DirectiveAT(value=directive_value)
            if directive.valid:
                output_filename = generic_temp_folder / directive.args[0]
            else:
                output_filename = generic_temp_folder / directive_value
            store_cfg(output_filename, d)
            saved_cfgs.append(output_filename)

        yconf = XConfig(output_cfg_filename)
        yconf.save_to(output_cfg_filename2)

        yconf_reloaded = XConfig(output_cfg_filename2)

        with pytest.raises(NotImplementedError):
            yconf.save_to(str(output_cfg_filename2) + "#IMPOSSIBLE.EXTENSION")

        assert not DeepDiff(yconf.to_dict(), sample_dict)
        assert not DeepDiff(yconf_reloaded.to_dict(), sample_dict)
        assert not DeepDiff(XConfig.from_dict(yconf_reloaded.to_dict()).to_dict(), yconf_reloaded.to_dict())
        assert len(XConfig.from_dict(yconf_reloaded.to_dict())) > len(sample_dict)  # YConf contains 2 more private keys!

        # remove last cfg file
        saved_cfgs[0].unlink()

        with pytest.raises(OSError):
            yconf = XConfig(output_cfg_filename)

    @pytest.mark.parametrize("cfg_extension", TESTABLE_ESTENSIONS)
    @pytest.mark.parametrize("data", data_to_test())
    def test_creation_root_replace(self, generic_temp_folder, data, cfg_extension):
        """ Test with double @@ -> replace value with content of content
        """
        generic_temp_folder = Path(generic_temp_folder)

        sample_dict = data['data']
        to_be_raplaced_keys = data['to_be_replaced']  # , _, _, to_be_raplaced_keys = data
        volatile_dict = copy.deepcopy(sample_dict)

        subtitutions_values = {}
        for k in to_be_raplaced_keys:
            random_name = k + f".{cfg_extension}"
            random_name = DirectiveAT.generate_directive_string('import_root', [random_name])
            subtitutions_values[random_name] = pydash.get(volatile_dict, k)

            pydash.set_(volatile_dict, k, random_name)

        output_cfg_filename = generic_temp_folder / f'out_config.{cfg_extension}'

        subtitutions_values[str(output_cfg_filename)] = volatile_dict

        saved_cfgs = []
        for directive_value, d in subtitutions_values.items():
            directive = DirectiveAT(value=directive_value)
            if directive.valid:
                output_filename = generic_temp_folder / directive.args[0]
            else:
                output_filename = generic_temp_folder / directive_value
            store_cfg(output_filename, d)
            saved_cfgs.append(output_filename)

        yconf = XConfig(output_cfg_filename)

        assert DeepDiff(yconf.to_dict(), sample_dict)


class TestXConfigReplace(object):

    @pytest.mark.parametrize("data", data_to_test())
    def test_replace(self, generic_temp_folder, data):

        sample_dict = data['data']
        placeholders = data['placeholders']  # , _, _, to_be_raplaced_keys = data
        environment_variables = data['environment_variables']

        conf = XConfig.from_dict(sample_dict)

        np.random.seed(66)

        to_replace = {}
        for p in placeholders:
            _p = Placeholder.from_string(p)
            to_replace[_p.name] = np.random.randint(0, 10)

        assert len(conf.available_placeholders()) == len(placeholders) + len(environment_variables)

        conf.check_available_placeholders(close_app=False)

        if len(conf.available_placeholders()) > 0:

            for name, p in conf.available_placeholders().items():
                assert isinstance(name, str)
                assert isinstance(p, Placeholder)

            with pytest.raises(SystemExit):
                conf.check_available_placeholders(close_app=True)
            conf.replace_variables_map(to_replace)

            chunks = conf.chunks_as_lists()
            for key, value in chunks:
                if not isinstance(value, Box) and not isinstance(value, BoxList):
                    assert value not in to_replace.keys()

            assert len(conf.available_placeholders()) == len(environment_variables)


class TestXConfigEnvironmentVariable(object):

    @pytest.mark.parametrize("data", data_to_test())
    def test_env_variables(self, generic_temp_folder, data):

        sample_dict = data['data']
        placeholders = data['placeholders']  # , _, _, to_be_raplaced_keys = data
        environment_variables = data['environment_variables']

        conf = XConfig.from_dict(sample_dict)
        np.random.seed(66)

        assert len(conf.available_placeholders()) == len(placeholders) + len(environment_variables)

        for envv in environment_variables:
            pl = Placeholder.from_string(envv)
            assert pl.name not in os.environ, f"this PYTEST needs no environment variable with name '{pl.name}' "

        # Load and parse environment variables but nobody set them!
        conf = XConfig.from_dict(sample_dict, replace_environment_variables=True)
        assert len(conf.available_placeholders()) == len(placeholders) + len(environment_variables)

        # Load and parse environment variables with manual set
        for envv in environment_variables:
            pl = Placeholder.from_string(envv)
            os.environ[pl.name] = str(np.random.uniform(0, 1, (1,)))

        conf = XConfig.from_dict(sample_dict, replace_environment_variables=True)

        import rich
        rich.print(conf)
        assert len(conf.available_placeholders()) == len(placeholders)

        for envv in environment_variables:
            pl = Placeholder.from_string(envv)
            del os.environ[pl.name]


class TestXConfigValidate(object):

    @pytest.mark.parametrize("data", data_to_test())
    def test_validation(self, generic_temp_folder, data):

        sample_dict = data['data']
        schema = data['schema']  # , _, _, to_be_raplaced_keys = data

        conf = XConfig.from_dict(sample_dict)
        conf.set_schema(schema)

        print(type(conf.get_schema()))
        conf.validate()
        assert conf.is_valid()

        conf.set_schema(None)
        assert conf.is_valid()

        invalid_schema = Schema({'fake': int})
        conf.set_schema(invalid_schema)
        with pytest.raises(SchemaMissingKeyError):
            conf.validate()
        assert not conf.is_valid()


class TestXConfigCopy(object):

    def test_copy(self, sample_configurations_data):

        for config in sample_configurations_data:
            filename = config['filename']
            xcfg = XConfig(filename=filename, no_deep_parse=True)
            print("#"*10)
            print(filename)
            print(xcfg)
            xcfg_copy = xcfg.copy()
            xcfg_copy.deep_parse()
            xcfg_correct = XConfig(filename=filename)
            assert not DeepDiff(xcfg_correct.to_dict(), xcfg_copy.to_dict())
            assert xcfg_copy == xcfg_correct


class TestXConfigChunksView(object):

    def test_chunks_view(self, sample_configurations_data):

        for config in sample_configurations_data:
            filename = config['filename']
            xcfg = XConfig(filename=filename, no_deep_parse=True)

            assert len(xcfg.chunks_as_lists()) > 0
            assert len(xcfg.chunks_as_lists()) == len(xcfg.chunks_as_tuples())
            assert len(xcfg.chunks()) == len(xcfg.chunks_as_tuples())

            ls = xcfg.chunks_as_lists()
            ts = xcfg.chunks_as_tuples()

            for idx in range(len(ls)):
                l, _ = ls[idx]
                t, _ = ts[idx]
                for kidx in range(len(l)):
                    assert l[kidx] == t[kidx]


class TestXConfigDeepSet(object):

    def test_deepset(self):

        cfg = {
            'l1': 'L1',
            'l2': {
                'l2_0': 'L20',
                'l2_1': 'L21',
            },
            'l3': {
                'l3_0': {
                    'l3_0_0': 'L300',
                    'l3_0_1': 'L301',
                    'l3_0_2': 'L302',
                }
            }
        }

        xcfg = XConfig.from_dict(cfg)

        for k, old_value in xcfg.chunks_as_lists():
            row_dict = {}
            pydash.set_(row_dict, k, old_value)

            xrow_dict = XConfig.from_dict(row_dict)
            for newk, newv in xrow_dict.chunks_as_lists():

                # VALID KEYS
                # Try to replace present keys only with ONLY_VALID_KEYS = TRUE, should be equal!
                new_xcfg = xcfg.copy()
                new_xcfg.deep_set(newk, newv, only_valid_keys=True)
                assert not DeepDiff(xcfg.to_dict(), new_xcfg.to_dict())

                # Try to replace present keys only with ONLY_VALID_KEYS = FALSE, should be equal!
                new_xcfg = xcfg.copy()
                new_xcfg.deep_set(newk, newv, only_valid_keys=False)
                assert not DeepDiff(xcfg.to_dict(), new_xcfg.to_dict())

                # INVALID KEYS
                new_keys = ['xxIMPOSSIBLE_KEY'] * 10
                # Try to replace not present keys but ONLY_VALID_KEYS = TRUE, should be equal!
                new_xcfg = xcfg.copy()
                new_xcfg.deep_set(new_keys, newv, only_valid_keys=True)
                assert not DeepDiff(xcfg.to_dict(), new_xcfg.to_dict())

                # Try to replace not present keys but ONLY_VALID_KEYS = False, should be different!
                new_xcfg = xcfg.copy()
                new_xcfg.deep_set(new_keys, newv, only_valid_keys=False)
                assert DeepDiff(xcfg.to_dict(), new_xcfg.to_dict())

    def test_deep_update(self):

        cfg = {
            'l1': 'L1',
            'l2': {
                'l2_0': 'L20',
                'l2_1': 'L21',
            },
            'l3': {
                'l3_0': {
                    'l3_0_0': 'L300',
                    'l3_0_1': 'L301',
                    'l3_0_2': 'L302',
                }
            }
        }

        cfg_to_replace = {
            'l3': {
                'l3_0': {
                    'l3_0_2': 'NEW_L302',
                }
            },
            'not_present_key': {
                'depth': {
                    'alpha': 'alpha'
                }
            }
        }

        xcfg = XConfig.from_dict(cfg)
        xcfg_to_replace = XConfig.from_dict(cfg_to_replace)

        # No Full merge
        new_cfg = xcfg.copy()
        new_cfg.deep_update(xcfg_to_replace, full_merge=False)
        assert len(xcfg.chunks_as_lists()) == len(new_cfg.chunks_as_lists())
        assert DeepDiff(xcfg.to_dict(), new_cfg.to_dict())

        # Full merge
        new_cfg = xcfg.copy()
        new_cfg.deep_update(xcfg_to_replace, full_merge=True)
        assert len(xcfg.chunks_as_lists()) < len(new_cfg.chunks_as_lists())


class TestXConfigValidateWithReplace(object):

    def test_validate_with_replace(self):

        for replace in [True, False]:
            cfg = {
                'one': {
                    's0': '1',
                    's1': '22.2',
                    's2': 'True'
                }
            }

            conf = XConfig.from_dict(cfg)

            schema = Schema({
                'one': {
                    's0': Use(int),
                    's1': Use(float),
                    's2': Use(bool),
                }
            })
            conf.set_schema(schema)
            conf.validate(replace=replace)

            assert isinstance(conf.one.s0, int) == replace
            assert isinstance(conf.one.s1, float) == replace
            assert isinstance(conf.one.s2, bool) == replace



from typing import Any, Dict, Sequence, Union
import inquirer
from choixe.configurations import XConfig
from choixe.placeholders import Placeholder, PlaceholderType


class XInquirer(object):

    @classmethod
    def safe_cast_placeholder(cls, value: Any, placeholder_type: PlaceholderType) -> bool:
        """ Checks if the value can be casted to the specified type

        :param value: any value
        :type value: Any
        :param placeholder_type: the placeholder type to check
        :type placeholder_type: PlaceholderType
        :return: either the value can be casted to the specified type or not
        :rtype: bool
        """

        try:
            if placeholder_type == PlaceholderType.BOOL:
                return value is True or value is False
            else:
                PlaceholderType.cast(value, placeholder_type)
            return True
        except Exception as e:
            print(e)
            return False

    @classmethod
    def placeholder_to_question(cls, placeholder: Placeholder) -> Union[inquirer.Text, inquirer.List]:
        """ Converts a placeholder to an inquirer question (to List if the placeholder has
        options or is boolean, Text otherwise)

        :param placeholder: the placeholder
        :type placeholder: Placeholder
        :return: the inquirer question
        :rtype: Union[inquirer.Text, inquirer.List]
        """

        message = f'{placeholder.name} ({placeholder.plain_type})'
        def validation(_, x): return cls.safe_cast_placeholder(x, placeholder.type)

        if len(placeholder.options) > 0 or placeholder.type == PlaceholderType.BOOL:

            if placeholder.type == PlaceholderType.BOOL:
                choices = [True, False]
                default = placeholder.default_value
                if default is not None:
                    default = True if placeholder.default_value.lower() == 'true' else False
            else:
                choices = placeholder.options
                default = placeholder.default_value

            question = inquirer.List(placeholder.name,
                                     message=message,
                                     choices=choices,
                                     default=default,
                                     validate=validation)
        else:
            question = inquirer.Text(placeholder.name,
                                     message=message,
                                     default=placeholder.default_value,
                                     validate=validation)
        return question

    @classmethod
    def unique_placeholders_with_order(cls, placeholders: Dict[str, Placeholder]) -> Sequence[Placeholder]:
        """ Gets uniques placeholders from an xconfig mantaining the order

        :param placeholders: dict of placeholders (the output of XConfig.available_placeholders())
        :type placeholders: Dict[str, Placeholder]
        :return: the unique placeholders
        :rtype: Sequence[Placeholder]
        """

        unique_names = set([ph.name for k, ph in placeholders.items()])
        unique_phs = []
        for k, ph in placeholders.items():
            if ph.name in unique_names:
                unique_phs.append(ph)
                unique_names.remove(ph.name)
        return unique_phs

    @classmethod
    def _system_prompt(cls, questions: Sequence[Union[inquirer.Text, inquirer.List]]) -> dict:
        return inquirer.prompt(questions)

    @classmethod
    def prompt(cls, xconfig: XConfig, close_app: bool = True) -> XConfig:
        """ Prompts the placeholders of an xconfig and fill them
        with the user answers, it returns a copy of the original xconfig

        :param xconfig: the xconfig
        :type xconfig: XConfig
        :param close_app: if True the app will be closed if all the placeholder are not filled, defaults to True
        :type close_app: bool, optional
        :return: the filled xconfig
        :rtype: XConfig
        """

        # copies the xconfig
        xconfig = XConfig.from_dict(xconfig.to_dict())
        # gets unique placeholders preserving order
        unique_phs = cls.unique_placeholders_with_order(xconfig.available_placeholders())
        # user interaction
        questions = [cls.placeholder_to_question(x) for x in unique_phs]
        answers = cls._system_prompt(questions)
        print("ANSW", answers)
        # replaces placeholders with inquirer answers
        xconfig.replace_variables_map(answers)
        xconfig.check_available_placeholders(close_app=close_app)
        return xconfig



import click


@click.command('check', help='Check Configuration file')
@click.option('-c', '--configuration_file', required=True, help="Input configuration file.")
@click.option('--close/--noclose', default=True, help="Raises system error if placeholders found")
def check(configuration_file, close):

    from choixe.configurations import XConfig
    import rich

    try:
        cfg = XConfig(filename=configuration_file)
        rich.print(cfg.to_dict())
    except Exception as e:
        rich.print(f"[red]Invalid configuration file: {e}[/red]")
        import sys
        sys.exit(1)

    rich.print("[green]Configuration is good![/green]")

    # d = cfg.to_dict(discard_private_qualifiers=False)
    cfg.check_available_placeholders(close_app=close)





from enum import Enum, auto
from typing import Union
from choixe.directives import Directive, DirectiveConsumer, DirectiveFactory
from choixe.importers import Importer, ImporterType


class PlaceholderType(Enum):
    INT = auto()
    STR = auto()
    FLOAT = auto()
    BOOL = auto()
    PATH = auto()
    DATE = auto()
    ENV = auto()
    OBJECT = auto()
    CFG = auto()
    CFG_ROOT = auto()

    @classmethod
    def values(cls):
        return list([c.name.lower() for c in cls])

    @classmethod
    def get_type(cls, value: str) -> Union['PlaceholderType', None]:
        c = [x for x in value.split('(') if len(x) > 0]
        if len(c) > 0:
            v = c[0]
            return PlaceholderType[v.upper()]
        return None

    @classmethod
    def get_type_as_string(cls, value: str) -> str:
        tp = cls.get_type(value)
        if isinstance(tp, PlaceholderType):
            return tp.name
        return 'None'

    @classmethod
    def cast(cls, value: any, tp: 'PlaceholderType'):
        if tp == PlaceholderType.BOOL:
            return bool(value)
        elif tp == PlaceholderType.STR:
            return str(value)
        elif tp == PlaceholderType.INT:
            return int(value)
        elif tp == PlaceholderType.FLOAT:
            return float(value)
        elif tp == PlaceholderType.PATH:
            return value
        elif tp == PlaceholderType.DATE:
            return value
        elif tp == PlaceholderType.ENV:
            return str(value)
        elif tp == PlaceholderType.OBJECT:
            return value
        elif tp == PlaceholderType.CFG:
            return Importer.generate_importer_directive(ImporterType.IMPORT, value)
        elif tp == PlaceholderType.CFG_ROOT:
            return Importer.generate_importer_directive(ImporterType.IMPORT_ROOT, value)
        else:
            raise NotImplementedError(f'No cast for type [{tp}] on [{value}]')


class Placeholder(DirectiveConsumer):

    def __init__(self, directive: Directive):
        super().__init__(directive=directive)

    def is_valid(self):
        if self._directive.valid:
            return self._directive.label in PlaceholderType.values()
        return False

    @property
    def name(self):
        if self.is_valid():
            if len(self._directive.args) > 0:
                return self._directive.args[0]
        return None

    @property
    def options(self):
        if self.is_valid():
            return self._directive.args[1:]
        return []

    @property
    def default_value(self):
        if self.is_valid():
            return self._directive.default_value
        return None

    @property
    def plain_type(self):
        if self.is_valid():
            return PlaceholderType.get_type_as_string(self._directive.label)
        return PlaceholderType.get_type_as_string('')

    @property
    def type(self):
        if self.is_valid():
            return PlaceholderType.get_type(self._directive.label)
        return None

    def cast(self, value: any) -> any:
        return PlaceholderType.cast(value, self.type)

    @classmethod
    def from_string(cls, value: str) -> 'Placeholder':
        directive = DirectiveFactory.build_directive_from_string(value)
        if directive:
            return Placeholder(directive=directive)
        return None



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('requirements.txt') as requirements_file:
    requirements = requirements_file.readlines()

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest>=3', ]

setup(
    author="daniele de gregorio",
    author_email='daniele.degregorio@eyecan.ai',
    python_requires='>=3.5',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="Configurations Superpowers",
    entry_points={
        'console_scripts': [
            'choixe=choixe.cli.main:choixe',
        ],
    },
    install_requires=requirements,
    license="GNU General Public License v3",
    long_description=readme,
    include_package_data=True,
    keywords='choixe',
    name='choixe',
    packages=find_packages(include=['choixe', 'choixe.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/eyecan-ai/choixe',
    version='0.1.8',
    zip_safe=False,
)



import pytest
from choixe.importers import Importer
from choixe.configurations import XConfig
from choixe.placeholders import Placeholder


class TestPlaceholders:

    def test_directiveat_defaults(self):

        items = [
            {
                'values': ['VNAME', 4, 666, 1222],
                'template':'@int({},default={})'
            },
            {
                'values': ['ANAME', 4.1, 666.6, 1.222, 66.6666],
                'template':'@float({},default={})'
            },
            {
                'values': ['SNAME', 'bb', 'ccc'],
                'template':'@str({},default={})'
            }
        ]

        for item in items:
            values = item['values']
            template = item['template']
            for v in values[1:]:
                t = template.format(','.join(map(str, values)), v)
                print("Template, ", t)
                p = Placeholder.from_string(t)
                assert p is not None
                assert p.cast(p.default_value) == v

                assert len(p.options) == len(values) - 1
                print(p.directive._kwargs)


class TestCFGPlaceholders:

    def test_cfg_placeholder(self, sample_configurations_data):

        for sample_cfg in sample_configurations_data:

            fname = sample_cfg["filename"]
            root_cfg = {
                'first_level': 1,
                'second': {
                    'second.1': f'@cfg(TO_IMPORT,{fname})',
                    'second.2': f'@cfg_root(TO_IMPORT_ROOT,{fname})',
                }
            }
            print(root_cfg)

            root_cfg = XConfig.from_dict(root_cfg)
            copy_cfg = root_cfg.copy()

            assert len(root_cfg.available_placeholders()) == 2

            # IMPORT
            for key, placeholder in root_cfg.available_placeholders().items():
                assert placeholder is not None
                assert len(placeholder.options) == 1
                importer_string = placeholder.cast(placeholder.options[0])
                print(importer_string)
                importer = Importer.from_string(importer_string)
                assert importer is not None
                print("REPLACING", key)
                root_cfg.replace_variable(placeholder.name, placeholder.options[0])

            assert len(root_cfg.available_placeholders()) == 0

            print("\nNEW_CFG", root_cfg)
            root_cfg.deep_parse()
            print("\nNEW_CFG", root_cfg)

            new_keys = [x[0] for x in root_cfg.chunks_as_tuples(discard_private_qualifiers=True)]
            old_keys = [x[0] for x in copy_cfg.chunks_as_tuples(discard_private_qualifiers=True)]

            assert len(new_keys) > len(old_keys)
